'''
Applications of the SMAP Level 4 Carbon (L4C) model, including sub-modules
related to L4C model calibration (`pyl4c.apps.calibration`) and forward run
simulation (`pyl4c.apps.l4c`). **Notably, down-scaling of L4C data from 9-km
resolution to 1-km resolution can be performed in with
`pyl4c.apps.resample`.**
'''
